#ifndef SUNOS
#include <sys/systeminfo.h>
#endif
#include <sys/types.h>
#include <time.h>
#include "xkernel.h"
#include "udp.h"
#include "ip.h"
#include "x_stdio.h"

#define argPrefix(str) ((!strncmp(arg, str, strlen(str))) && strlen(arg)>strlen(str))
#define argEq(str)     (!strcmp(arg, str))
#define HOST_TYPE      IPhost

#ifndef STR2HOST
#define STR2HOST str2ipHost
#endif

#define TRACE_VAR traceudptestp
#define PROT_STRING "udp"
#define TIMES 5
#define DELAY 3
#define SPIN_DELAY 1
#define FAILURE_LIMIT 3

typedef struct {
    Sessn    lls;
    Sessn    controlSessn;
    Msg      savedMsg;
    XkHandle clientPushResult;
    XkHandle serverPushResult;
    int      done, idle;
    int      clientReceived;
    XTime    starttime, endtime;
} PState;

static HOST_TYPE ServerAddr = { 0, 0, 0, 0 };
static HOST_TYPE ClientAddr = { 0, 0, 0, 0 };
static HOST_TYPE myHost;
static long      serverPort = 1789;

static int  serverParam, clientParam;
static char *serverString;

/* Sizes of messages sent, in bytes */
#define NUM_TESTS 11
static int size[] = { 1, 200, 400, 600, 800, 1024, 2048, 4096, 8192, 16384, 32768 };
static int trips = 100;

int TRACE_VAR;

static void     client(Event, void *);
static void     server(Event, void *);
static XkReturn clientDemux(Protl, Sessn, Msg *);
static int      runTest(Protl, int, int);
static XkReturn serverDemux(Protl, Sessn, Msg *);
static void     processOptions(void);
static XkReturn closeDone(Sessn);
static XkReturn customOpenDone(Protl, Protl, Sessn, Protl);

int udptest_init(Protl);

int udptest_init(Protl self)
{
    Protl  llp;
    PState *ps;

    processOptions();

    printf("%s timing test\n", PROT_STRING);
    llp = xGetProtlDown(self, 0);
    if (!xIsProtl(llp))
        Kabort("Test protocol has no lower protocol");
    xControlProtl(xGetProtlDown(self, 0), GETMYHOST, (char *)&myHost,
		  sizeof(HOST_TYPE));
    ps = X_NEW(PState);
    bzero((char *)ps, sizeof(PState));
    self->state = (VOID *)ps;

    if (serverParam)
        evDetach(evSchedule(server, self, 0));
    else if (clientParam) {
        STR2HOST(&ServerAddr, serverString);
        ClientAddr = myHost;
        evDetach(evSchedule(client, self, 0));
    }
    return 0;
}

static void
processOptions()
{
    int  i;
    char *arg;

    for (i = 1; i < globalArgc; i++) {
        arg = globalArgv[i];
        if (argEq("-s"))
	    serverParam = 1;
        else if (argPrefix("-c")) {
	    clientParam = 1;
	    serverString = arg + 2;
	}
        else if (argEq("-c")) {
	    clientParam = 1;
	    serverString = globalArgv[i+1];
	    i++;
        }
        else if (argPrefix("-trips="))
	    sscanf(arg + strlen("-trips="), "%d", &trips);
        else if (argEq ("-trips")) {
	    sscanf(globalArgv[i+1], "%d", &trips);
	    i++;
	}
    }
}

/*
 *  Server startup routine
 */
static void
server(Event ev, VOID *foo)
{
    Part   p;
    Protl  myProtl = (Protl)foo;
    char   buf[256];
    time_t t = time(0);

    printf("%s", ctime(&t));
#ifdef SUNOS
    gethostname(buf, 255);
#else
    sysinfo(SI_HOSTNAME, buf, 255);
#endif
    printf("I am the server, running on %s\n\n", buf);
    xAssert(xIsProtl(myProtl));

    /* Since we are a test protocol, need to fill out own method table. */
    myProtl->demux = serverDemux;
    myProtl->opendone = customOpenDone;
    myProtl->closedone = closeDone;

    partInit(&p, 1);
    partPush(p, ANY_HOST, 0);
    partPush(p, &serverPort, sizeof(long));

    if (xOpenEnable(myProtl, myProtl, xGetProtlDown(myProtl, 0), &p) ==
	XK_FAILURE) {
        printf("%s test server can't openenable lower protocol\n",
	       PROT_STRING);
    }
    else
        printf("%s test server done with xopenenable\n", PROT_STRING);
}

/*
 * Client initialization routine
 */
static void
client(Event ev, VOID *foo)
{
    Protl  myProtl = (Protl)foo;
    PState *ps;
    char   buf[256];
    time_t t = time(0);
    int    i;
    Part   p[2];

    printf("%s", ctime(&t));
#ifdef SUNOS
    gethostname(buf, 255);
#else
    sysinfo(SI_HOSTNAME, buf, 255);
#endif
    printf("I am the client, running on %s\n", buf);
    printf("\nRunning UDP ping-pong test\n");
    printf("Each message makes %d round trips\n\n", trips);

    xAssert(xIsProtl(myProtl));
    ps = (PState *)myProtl->state;

    /* Since we are a test protocol, need to fill out own method table. */
    myProtl->demux = clientDemux;

    partInit(p, 1);
    partPush(p[0], &ServerAddr, sizeof(IPhost));
    partPush(p[0], &serverPort, sizeof(long));

    ps->lls = xOpen(myProtl, myProtl, xGetProtlDown(myProtl, 0), p);
    if (ps->lls == ERR_SESSN) {
        printf("%s test: open failed!\n", PROT_STRING);
        Kabort("End of test");
    }

    customOpenDone(myProtl, xGetProtlDown(myProtl, 0), ps->lls, myProtl);

    for (i = 0; i < NUM_TESTS; i++)
        runTest(myProtl, size[i], i);

    xClose(ps->lls);

    fprintf(stderr, "End of test\n");
}

static
#if defined(__GNUC__)
inline
#endif
void checkHandle(h, str)
XkHandle h;
char     *str;
{
    switch (h) {
        case XMSG_ERR_HANDLE:
        case XMSG_ERR_WOULDBLOCK:
	    sprintf(errBuf, "%s returns error handle %d", str, h);
	    Kabort(errBuf);
        default:
	    break;
    }
}

int runTest(self, len, testNumber)
Protl self;
int   len;
int   testNumber;
{
    Msg    msg;
    PState *ps = (PState *)self->state;
    XTime  total;
    int    test, failures;

    xAssert(ps);
    msgConstructAllocate(&ps->savedMsg, len);

    for (test = 0; test < TIMES; test++) {
        printf("Test %d, run %d ...\n", testNumber, test);
        printf("%d trips using msg length: %d\n", trips, len);

        ps->clientReceived = 0;
        ps->done = 0;
        failures = 0;
        xGetTime(&(ps->starttime));

        msgConstructCopy(&msg, &ps->savedMsg);
        ps->clientPushResult = xPush(ps->lls, &msg);
        checkHandle(ps->clientPushResult, "client push");
        msgDestroy(&msg);

        /* Wait until test is completed. */
        do {
	    ps->idle = 1;
	    Delay(SPIN_DELAY * 1000);
	    if (ps->idle) {
	        failures++;
	        if (failures >= FAILURE_LIMIT)
		    break;
	    }
	} while (!ps->done);

        if (!ps->done) {
	    printf("\nTest failed after %d trips... terminating\n\n",
		   ps->clientReceived);
	    continue;
	}

        xSubTime(&total, ps->endtime, ps->starttime);
        printf("\nTime for test %d, run %d: %6d.%06d\n\n", testNumber, test,
	       total.sec, total.usec);

        Delay(DELAY * 1000);
    }

    msgDestroy(&ps->savedMsg);

    return 0;
}

XkReturn
serverDemux(self, lls, dg)
Protl self;
Sessn lls;
Msg   *dg;
{
    PState *ps = (PState *)self->state;
    static int c = 1;

    xIfTrace(udptestp, TR_MAJOR_EVENTS) {
        putchar('.');
        if (!(c++ % 50))
	    putchar('\n');
    }

    /* Echo packet back to client */
    ps->serverPushResult = xPush(lls, dg);
    checkHandle(ps->serverPushResult, "server push");

    return XK_SUCCESS;
}

static XkReturn
clientDemux(self, lls, dg)
Protl self;
Sessn lls;
Msg   *dg;
{
    PState *ps = (PState *)self->state;
    Msg    tmpMsg;
    static int c = 1;

    ps->clientReceived++;
    ps->idle = 0;
    xIfTrace(udptestp, TR_MAJOR_EVENTS) {
        putchar('.');
        if (!(c++ % 50))
	    putchar('\n');
    }

    if (ps->clientReceived < trips) {
        /*
         * We need to construct a copy of the original message rather than
	 * just loop back the incoming message to avoid an increasingly
	 * fragmented message structure in the case of loopback.
         */

        msgConstructCopy(&tmpMsg, &ps->savedMsg);
        ps->clientPushResult = xPush(lls, &tmpMsg);

        checkHandle(ps->clientPushResult, "client push");
        msgDestroy(&tmpMsg);
    }
    else {
        xGetTime(&ps->endtime);
        ps->done = 1;
    }
    return XK_SUCCESS;
}

static XkReturn
closeDone(lls)
Sessn lls;
{
    xTrace2(udptestp, TR_MAJOR_EVENTS, "\n%s test -- closedone (%x) called",
	    PROT_STRING, lls);
    if (serverParam)
        xClose(lls);
    return XK_SUCCESS;
}

static XkReturn
customOpenDone(self, llp, s, hlpType)
Protl self, llp, hlpType;
Sessn s;
{
    xTrace0(udptestp, TR_MAJOR_EVENTS, "udp test program openDone");
    if (serverParam)
        xDuplicate(s);
    return XK_SUCCESS;
}
