/*
 * @COPYRIGHT@
 * 
 * x-kernel v3.3
 * 
 * Copyright (c) 1993,1991,1990,1996  Arizona Board of Regents
 * 
 * @COPYRIGHT@
 *
 * $RCSfile: scout_malloc.c,v $
 *
 * HISTORY
 * $Log: scout_malloc.c,v $
 * Revision 1.11  1996/01/30  22:49:51  slm
 * Updated copyright and version.
 *
 * Revision 1.10  1995/02/24  02:12:14  davidm
 * When determining rounded-size, it is necessary to ensure the allocated
 * block is at least as long as a struct flist (not just struct ulist!).
 *
 * Revision 1.9  1995/02/08  23:45:04  davidm
 * Improved scout_panic() message.
 *
 * Revision 1.8  1994/11/20  01:42:15  davidm
 * Added support to keep track of allocated memory (SCOUT_DEBUG only).
 *
 * Revision 1.7  1994/08/28  03:14:14  davidm
 * Added sentinels to make freeing memory faster.  Yielded an 8 us improvement
 * for TCP test!
 *
 * Revision 1.6  1994/08/28  02:05:58  davidm
 * Added some predictions, removed more_core() and made other
 * simplifications.
 *
 * Revision 1.5  1994/08/15  02:06:55  davidm
 * Minor changes.
 *
 * Revision 1.4  1994/07/13  20:54:55  dpw
 * Changed to prevent size reduction of old blocks.
 *
 * Revision 1.3  1994/07/13  01:29:55  dpw
 * Added scout_realloc() routine, changed scout_free() to return gracefully
 * when passed a null pointer
 *
 * Revision 1.2  1994/06/11  00:02:31  menze
 * Added copyright notice
 *
 * Revision 1.1  1994/04/01  22:54:07  davidm
 * Initial revision
 */
#include "scout.h"
#include "scout_malloc.h"
#include "scout_trace.h"
#include "scout_util.h"

#ifndef MAX
# define MAX(a,b)	(((a)>(b))?(a):(b))
#endif

/*
 * Manages memory with a doubly linked, cyclic free list.
 *	- search starts where previous search ended
 *	- contiguous blocks are coalesced on free()
 *	- free list contains memory blocks in order of increasing addresses
 */
#ifdef SCOUT_DEBUG
 /*
  * A block has magic set to MALLOC_MAGIC if it's free and
  * MALLOC_MAGIC+1 if it's allocated.
  */
# define MALLOC_MAGIC	0xC0FFEE
#endif /* SCOUT_DEBUG */

int trace_scout_malloc = 0;

/*
 * Free-list header:
 */
struct flist {
    unsigned long size;		/* size of this block (includes header size) */
    struct flist *next;		/* next free block */
    struct flist *prev;		/* previous free block */
#ifdef SCOUT_DEBUG
    unsigned int  magic;
#endif /* SCOUT_DEBUG */
};

/*
 * In-use list header.  Normally, all we need to know is the size of a
 * memory block, but with debugging turned on, we need to keep more
 * state.  The size field *must* be common with the size field in
 * struct flist!
 */
struct ulist {
    unsigned long size;		/* size of this block (includes header size) */
#ifdef SCOUT_DEBUG
    struct ulist *next;		/* next free block */
    unsigned long owner;
    unsigned int  magic;
#endif /* SCOUT_DEBUG */
};

static struct flist *free_list = 0;
static struct flist *low_sentinel = (struct flist*) -1L;
static struct flist *high_sentinel = 0;
#ifdef SCOUT_DEBUG
static struct ulist in_use_list = {0, 0};
#endif /* SCOUT_DEBUG */

#define MINSIZE	(sizeof(struct flist) + sizeof(int))


#ifdef SCOUT_DEBUG

void
dump_in_use_list(void)
{
    struct ulist *ul;
    int col = 0, new_line;
    long in_use = 0;

    for (ul = in_use_list.next; ul; ul = ul->next) {
	new_line = col++ & 1;
	printf("(%p+%-8ld %lx)%c", ul, ul->size, ul->owner,
	       new_line ? '\n' : ' ');
	in_use += ul->size;
    } /* for */
    printf("\n%ld bytes in use.\n", in_use);
} /* dump_in_use_list */


#endif /* SCOUT_DEBUG */

void
dump_free_list(void)
{
    struct flist *el = free_list;
    long free_space = 0;
    int len = 0;
    char bogus = 0;

    printf("FL: free_list @ %p:", free_list);
    if (!el) {
        printf("FL: <EMPTY>");
    } else {
        while (el->next > el) {
            el = el->next;
        } /* while */
        do {
	    if (el->next->prev != el) {
		bogus = 1;
		printf("\nFL: BOGUS: el=%p, el->next=%p, el->next->prev=%p\n",
		       el, el->next, el->next->prev);
	    } /* if */
	    ++len;
	    free_space += el->size;
            el = el->next;
#ifdef SCOUT_DEBUG
	    if (el->magic != MALLOC_MAGIC) {
		bogus = 1;
		printf("BOGUS:");
	    } /* if */
#endif
	    printf(" %p+%ld", el, el->size);
        } while (el->next > el);
    } /* if */
    printf("\nFL: free space=%ld\n", free_space);
    printf("FL: length of free list=%d\n", len);
    if (bogus) {
	scout_panic(__FILE__":dump_free_list: mangled free list!");
    } /* if */
} /* dump_free_list */


static void
insert_sentinel(struct flist *el)
{
    struct flist *p = free_list;

    el->size = 0;	/* fake size so insert() won't merge them away */
#ifdef SCOUT_DEBUG
    el->magic = MALLOC_MAGIC;
#endif /* SCOUT_DEBUG */

    if (!p PREDICT_FALSE) {
	/* list was empty: */
	el->next = el;
	el->prev = el;
	free_list = el;
    } else {
	/*
	 * search for appropriate location (the search starts potentially
	 * in the middle of the free list, that's why the conditional
	 * looks somewhat complex):
	 */
	while (!((p < el && el < p->next)
	       || (p >= p->next && (el > p || el < p->next))))
	{
	    p = p->next;
	} /* while */

	/* insert sentinel into free_list: */
	el->next = p->next;
	el->prev = p;
	p->next->prev = el;
	p->next = el;
    } /* if */
} /* insert_sentinel */


/*
 * EL is the block of memory to insert in the free list.
 * Returns the block of memory with which EL was coalesced
 * (or EL itself).
 */
static inline struct flist*
insert(struct flist *el)
{
    struct flist *p = free_list;
    register int need_insert = TRUE;

    if (!p PREDICT_FALSE) {
	/* list was empty: */
	el->next = el;
	el->prev = el;
	free_list = el;
    } else {
	/*
	 * Search for appropriate location.  Thanks to the sentinels,
	 * we always insert *between* two existing elements.  As
	 * the list could be longish, we make one condition true
	 * after the other.  However, after establishing "p>el"
	 * for the first time, establishing "el>p->next" could
	 * invalidate the former assertion, so we have to make sure
	 * we re-establish it.
	 */
	while (p > el) p = p->next;
	while (el > p->next) p = p->next;
	while (p > el) p = p->next;

	if ((char*)p + p->size == (char*)el && el->size) {
	    /* merge with predecessor block: */
#ifdef SCOUT_DEBUG
	    el->magic = 0;
#endif /* SCOUT_DEBUG */
	    p->size += el->size;
	    el = p;
	    need_insert = FALSE;
	} /* if */
	if ((char*)el + el->size == (char*)p->next && p->next->size) {
	    /* merge with successor block: */
#ifdef SCOUT_DEBUG
	    p->next->magic = 0;
#endif /* SCOUT_DEBUG */
	    el->size += p->next->size;
	    if (p->next == p PREDICT_FALSE) {
		/* take care of singleton list: */
		el->next = el;
		el->prev = el;
	    } else {
		el->next = p->next->next;
		el->next->prev = el;
		if (el != p) {
		    el->prev = p;
		    p->next = el;
		} /* if */
	    } /* if */
	    need_insert = FALSE;
	} /* if */
	if (need_insert) {
	    el->next = p->next;
	    el->prev = p;
	    p->next->prev = el;
	    p->next = el;
	} /* if */
    } /* if */
    return el;
} /* insert */


void*
scout_malloc(int size)
{
    struct flist *el;
    struct ulist *ul;
    int rsize;

    ScoutTrace(scout_malloc, TR_DETAILED, ("scout_malloc(size=%d)", size));

    /*
     * Up size by size of header and make it a multiple of sizeof(void*):
     */
    rsize = (MAX(size + sizeof(struct ulist), sizeof(struct flist))
	     + sizeof(void*) - 1) & ~(sizeof(void*) - 1);
    el = free_list;
    if (!el PREDICT_FALSE) {
	return 0;
    } /* if */

    /* walk free list and take first block that is large enough: */
    while (TRUE) {
#ifdef SCOUT_DEBUG
	if (el->magic != MALLOC_MAGIC) {
	    scout_abort("scout_malloc: bad magic: %#x instead of %#x",
			el->magic, MALLOC_MAGIC);
	} /* if */
#endif /* SCOUT_DEBUG */
	if (el->size >= rsize) {
	    break;
	} /* if */
	el = el->next;
	if (el == free_list PREDICT_FALSE) {
	    return 0;
	} /* if */
    } /* while */

    if (el->size - rsize >= MINSIZE) {
	/* start next search at reduced block: */
	free_list = el;

	/* reduce block (from end): */
	el->size -= rsize;
	el = (struct flist*) ((char*)el + el->size);
	el->size = rsize;
    } else {
	/* remove block completely from free-list: */
	if (el->next == el PREDICT_FALSE) {
	    /* free list is empty now: */
	    free_list = 0;
	} else {
	    el->prev->next = el->next;
	    el->next->prev = el->prev;

	    /* start next search after returned block: */
	    free_list = el->next;
	} /* if */
    } /* if */

    ul = (struct ulist*) el;

#ifdef SCOUT_DEBUG
    ul->magic = MALLOC_MAGIC + 1;
    {
	register unsigned long ra asm("26");
	struct ulist *p;

	ul->owner = ra;
	ul->next = in_use_list.next;
	in_use_list.next = ul;

	p = in_use_list.next;
	while (p) {
	    if (p->next && ((u_long)p->next >> 32) != 0xfffffc00) {
		scout_panic("scout_malloc: memory corruption!\n");
	    } /* if */
	    p = p->next;
	} /* while */
    }
#endif /* SCOUT_DEBUG */

    ScoutTrace(scout_malloc, TR_DETAILED,
	       ("scout_malloc: ptr=%p", (char*)el + sizeof(struct flist)));
    return (char*)ul + sizeof(struct ulist);
} /* scout_malloc */


void
scout_free(void *ptr)
{
    struct flist *el = (struct flist*) (ptr - sizeof(struct ulist));
#ifdef SCOUT_DEBUG
    struct ulist *ul = (struct ulist*) el;
    struct ulist *p = &in_use_list;
#endif

    ScoutTrace(scout_malloc, TR_DETAILED, ("scout_free(ptr=%p)", ptr));
    if (!ptr PREDICT_FALSE) {
	return;
    } /* if */
#ifdef SCOUT_DEBUG
    if (ul->magic != MALLOC_MAGIC + 1) {
	scout_abort("scout_free: bad magic: %#x instead of %#x",
		    ul->magic, MALLOC_MAGIC + 1);
    } /* if */
    el->magic = MALLOC_MAGIC;

    /* remove ul from in-use list: */
    while (p && p->next != ul) {
	p = p->next;
    } /* while */
    if (!p) {
	scout_panic("scout_free: freeing memory that's not in use!");
    } /* if */
    p->next = ul->next;
#endif /* SCOUT_DEBUG */

    free_list = insert(el);
    return;
} /* scout_free */


void*
scout_realloc(void *ptr, int newsize)
{
    struct ulist *ul = (struct ulist*) (ptr - sizeof(struct ulist));
    int rnewsize;

    ScoutTrace(scout_malloc, TR_DETAILED, ("scout_realloc(ptr=%p,newsize=%d)",
					   ptr, newsize));

    /*
     * Up size by size of header and make it a multiple of sizeof(void*):
     */
    rnewsize = (MAX(newsize + sizeof(struct ulist), sizeof(struct flist)) +
		+ sizeof(void*) - 1) & ~(sizeof(void*) - 1);

    /* free mem and return null pointer if size = 0: */
    if (newsize == 0) {
	scout_free(ptr);
	return 0;
    } /* if */

    /* malloc a new block if passed a null pointer: */
    if (ptr == 0) {
	return scout_malloc(newsize);
    } /* if */

#ifdef SCOUT_DEBUG
    if (ul->magic != MALLOC_MAGIC + 1) {
	scout_abort("scout_realloc: bad magic: %#x instead of %#x",
		    ul->magic, MALLOC_MAGIC + 1);
    } /* if */
#endif /* SCOUT_DEBUG */

    /* copy relevant data to new block if size is greater than old size: */
    if (rnewsize > ul->size) {
	void *newptr = scout_malloc(newsize);
	int len = ul->size - sizeof(struct ulist);

	bcopy(ptr, newptr, len);
	scout_free(ptr);
	ScoutTrace(scout_malloc, TR_DETAILED,
		   ("scout_realloc: newptr=%p", newptr));
	return newptr;
    } /* if */
    return ptr;
} /* scout_realloc */


/*
 * Allocate "size" bytes such that the returned starting address
 * is a multiple of "align":
 */
void*
scout_malloc_aligned(int size, int log_alignment)
{
    struct ulist *ul, *pul;
    struct flist *nel;
    int rsize, align, tot_size, waste;
    char *aligned_addr;

    align = 1 << log_alignment;
    ScoutTrace(scout_malloc, TR_DETAILED,
	       ("scout_malloc_aligned(size=%d,align=%d)", size, align));

    /*
     * Up size by size of header and make it a multiple of sizeof(void*):
     */
    rsize = (MAX(size + sizeof(struct ulist), sizeof(struct flist))
	     + sizeof(void*) - 1) & ~(sizeof(void*)-1);

    ul = scout_malloc(rsize + align);
    pul = ul - 1;
    tot_size = pul->size;
    aligned_addr = (char*)(((u_long)(ul + 1) + align - 1) & ~(align - 1));
    ul = (struct ulist*)aligned_addr - 1;

    pul->size = (char*) ul - (char*) pul;
    ul->size = rsize;
#ifdef SCOUT_DEBUG
    ul->magic = MALLOC_MAGIC + 1;	/* this memory is in use now */
    ul->owner = (u_long) scout_malloc_aligned;
    ul->next = in_use_list.next;
    in_use_list.next = ul;
#endif /* SCOUT_DEBUG */
    waste = tot_size - pul->size - ul->size;
    if (waste > MINSIZE) {
	nel = (struct flist*)((char*)ul + rsize);
	nel->size = waste;
#ifdef SCOUT_DEBUG
	nel->magic = MALLOC_MAGIC;
#endif
	insert(nel);		/* free space behind block */
    } else {
	ul->size += waste;	/* don't bother freeing rest */
    } /* if */
    scout_free(pul + 1);	/* free space in front of block */

    return aligned_addr;
} /* scout_malloc_aligned */


void
scout_malloc_expand (void *mem, int size)
{
    struct flist *el;

    ScoutTrace(scout_malloc, TR_EVENTS,
	       ("scout_malloc_expand: mem=%p, size=%d", mem, size));

    if (size < 3 * sizeof(struct flist)) {
	scout_panic("scout_malloc_expand: ridiculously small expansion");
    } /* if */

    /* first, update sentinels: */

    if ((u_long) mem < (u_long) low_sentinel) {
	/* move low sentinel: */
	if ((long) low_sentinel != -1L) {
	    /* fixup size: */
	    low_sentinel->size = sizeof(*low_sentinel);
	    insert(low_sentinel);
	} /* if */
	low_sentinel = (struct flist*) mem;
	insert_sentinel(low_sentinel);
	mem  += sizeof(*low_sentinel);
	size -= sizeof(*low_sentinel);
    } /* if */

    if ((u_long) mem > (u_long) high_sentinel) {
	/* move high sentinel: */
	if (high_sentinel) {
	    /* fixup size: */
	    high_sentinel->size = sizeof(*high_sentinel);
	    insert(high_sentinel);
	} /* if */
	high_sentinel = mem + (size - sizeof(*high_sentinel));
	insert_sentinel(high_sentinel);
	size -= sizeof(*high_sentinel);
    } /* if */

    el = (struct flist*) mem;
    el->size = size;
#ifdef SCOUT_DEBUG
    el->magic = MALLOC_MAGIC;
#endif /* SCOUT_DEBUG */
    insert(el);
} /* scout_malloc_expand */

			/*** end of scout_malloc.c ***/
